// This file is common to the XML Moo Database browser and Moo Canada's IDE.

// What XML object are we currently looking at?
var CurrentId = null;
// Various cached attributes of this object.
var CachedValues = {};

function changeObject(newId) {
  if (CurrentId != newId) {
    CurrentId = newId;
    CachedValues = {};
    treeUpdate();
    // Scroll to the top of every pane.
    for (var x in Tab.List) {
      Tab.List[x].contentNode.scrollTop = 0;
    }
  }
}

// Someone, somewhere thinks that something might have changed.
// Look up the currently displayed node and pass it to any interested tabs.
function treeUpdate() {
  var node = parent.parent.getXmlNodeById(CurrentId);
  for (var x in Tab.List) {
    if (Tab.List[x].updateObject) {
      Tab.List[x].updateObject(node);
    }
  }
}

if (Tab.List['object']) {
  Tab.List['object'].updateObject = function(xmlNode) {
    if (!xmlNode) {
      document.getElementById("content_object_yes").style.display = "none";
      document.getElementById("content_object_none").style.display = "block";
      return;
    }
    document.getElementById("content_object_none").style.display = "none";
    document.getElementById("content_object_yes").style.display = "block";

    // Flags
    var flags = xmlNode.getAttribute('FLAGS');
    if (CachedValues['flags'] !== flags) {
      CachedValues['flags'] = flags;
      if (flags) {
        flags = " " + flags + " ";
        flags = flags.replace(" r ", " readable ");
        flags = flags.replace(" w ", " writable ");
        flags = flags.replace(" f ", " fertile ");
      } else {
        flags = "";
      }
      replaceChildNodes(document.getElementById("object_flags"), document.createTextNode(flags));
    }
    
    var inheritanceTreeNode, locationTreeNode, value;

    // Owner
    var ownerNumber = xmlNode.getAttribute('OWNER').replace('obj', '#');
    inheritanceTreeNode = parent.parent.TreeFrame.treeGetOrCreateObject(ownerNumber, 'inheritance')
    value = inheritanceTreeNode.firstChild.title;
    if (CachedValues['owner'] !== value) {
      CachedValues['owner'] = value;
      replaceChildNodes(document.getElementById("object_owner"), buildObjectLink(inheritanceTreeNode));
    }

    // Name
    inheritanceTreeNode = parent.parent.TreeFrame.treeGetOrCreateObject(CurrentId.replace('obj', '#'), 'inheritance');
    locationTreeNode = parent.parent.TreeFrame.treeGetOrCreateObject(CurrentId.replace('obj', '#'), 'location');
    value = inheritanceTreeNode.firstChild.title;
    if (CachedValues['name'] !== value) {
      CachedValues['name'] = value;
      replaceChildNodes(document.getElementById("object_name"), document.createTextNode(value));
    }

    // Parent
    value = inheritanceTreeNode.parentNode && inheritanceTreeNode.parentNode.firstChild.title;
    if (CachedValues['parent'] !== value) {
      CachedValues['parent'] = value;
      replaceChildNodes(document.getElementById("object_parent"), buildObjectLink(inheritanceTreeNode.parentNode));
    }

    // Location
    value = locationTreeNode.parentNode && locationTreeNode.parentNode.firstChild.title;
    if (CachedValues['location'] !== value) {
      CachedValues['location'] = value;
      replaceChildNodes(document.getElementById("object_location"), buildObjectLink(locationTreeNode.parentNode));
    }

    // Contents
    var objectList = document.createElement("SPAN");
    for (var x = 1; x < locationTreeNode.childNodes.length; x++) {
      objectList.appendChild(buildObjectLink(locationTreeNode.childNodes[x]));
      if (x + 1 < locationTreeNode.childNodes.length) {
        objectList.appendChild(document.createTextNode(", "));
      }
    }
    if (objectList.firstChild == null) {
      objectList.appendChild(document.createTextNode("none"));
      objectList.className = "none";
    }
    value = objectList.innerHTML;
    if (CachedValues['contents'] !== value) {
      CachedValues['contents'] = value;
      replaceChildNodes(document.getElementById("object_contents"), objectList);
    }

    // Children
    objectList = document.createElement("SPAN");
    for (var x = 1; x < inheritanceTreeNode.childNodes.length; x++) {
      objectList.appendChild(buildObjectLink(inheritanceTreeNode.childNodes[x]));
      if (x + 1 < inheritanceTreeNode.childNodes.length) {
        objectList.appendChild(document.createTextNode(", "));
      }
    }
    if (objectList.firstChild == null) {
      objectList.appendChild(document.createTextNode("none"));
      objectList.className = "none";
    }
    value = objectList.innerHTML;
    if (CachedValues['children'] !== value) {
      CachedValues['children'] = value;
      replaceChildNodes(document.getElementById("object_children"), objectList);
    }
  }
}

if (Tab.List['verbs']) {
  Tab.List['verbs'].updateObject = function(xmlNode) {
    if (!xmlNode) {
      document.getElementById("content_verbs_yes").style.display = "none";
      document.getElementById("content_verbs_none").style.display = "block";
      return;
    }
    document.getElementById("content_verbs_none").style.display = "none";
    var contentDiv = document.getElementById("content_verbs_yes");
    contentDiv.style.display = "block";

    // Tear down all existing content.
    while (contentDiv.firstChild) {
      contentDiv.removeChild(contentDiv.firstChild);
    }

    var id = xmlNode.getAttribute('ID');
    var xmlLeaf = xmlNode;
    do {
      updateVerbTable(xmlNode, xmlLeaf);
      xmlNode = xmlNode.parentNode;
    } while(xmlNode.nodeName == "OBJECT")

    var firstList = document.getElementById("verbcontrol_" + id);
    if (firstList) {
      listSetFold(firstList, true);
    }
  }
}


if (Tab.List['properties']) {
  Tab.List['properties'].updateObject = function(xmlNode) {
    if (!xmlNode) {
      document.getElementById("content_properties_yes").style.display = "none";
      document.getElementById("content_properties_none").style.display = "block";
      return;
    }
    document.getElementById("content_properties_none").style.display = "none";
    var contentDiv = document.getElementById("content_properties_yes");
    contentDiv.style.display = "block";

    // Tear down all existing content.
    while (contentDiv.firstChild) {
      contentDiv.removeChild(contentDiv.firstChild);
    }

    var id = xmlNode.getAttribute('ID');
    // ancestors is an array of parent objects from the current one to the root.
    var ancestors = [];
    var node = xmlNode;
    do {
      ancestors.push(node);
      node = node.parentNode;
    } while(node.nodeName == "OBJECT")

    // propNames is a dictionary keyed by the property name with a value of:
    // [xmlNode, perms, owner, value]
    // Where xmlNode is the last object that prop wasn't clear on.
    var propNames = {};
    // propObjects is an array of arrays per ancestor.
    // Each sub array contains an array of property names defined on that object.
    var propObjects = [];
    // propObject is a sub array for propObjects.
    var propObject;
    // Populate these data structures.
    for (var x = ancestors.length - 1; x >= 0; x--) {
      propObject = [];
      parsePropObject(ancestors[x], propObject, propNames);
      propObjects[x] = propObject;
    }

    for (var x = 0; x < ancestors.length; x++) {
      updatePropTable(ancestors[x], xmlNode, propObjects[x], propNames);
    }
    var firstList = document.getElementById("propcontrol_" + id);
    if (firstList) {
      listSetFold(firstList, true);
    }
  }
}


function parsePropObject(xmlNode, propObject, propNames) {
  var childXmlNodeLength = xmlNode.childNodes.length;
  var childXmlNode;
  var perms, owner, pname, value;
  for (var x = 0; x < childXmlNodeLength; x++) {
    childXmlNode = xmlNode.childNodes[x];
    if (childXmlNode.nodeName == "PROPERTY") {
      perms = childXmlNode.getAttribute('PERMS') || "";
      owner = childXmlNode.getAttribute('OWNER');
      pname = "";
      value = ""; // Note: a Moo empty string will be "\"\""
      var grandchildXmlNodeLength = childXmlNode.childNodes.length;
      var grandchildXmlNode;
      for (var y = 0; y < grandchildXmlNodeLength; y++) {
        grandchildXmlNode = childXmlNode.childNodes[y];
        if (grandchildXmlNode.nodeName == "NAME" && grandchildXmlNode.firstChild) {
          pname += grandchildXmlNode.firstChild.nodeValue;
        }
        if (grandchildXmlNode.nodeName == "VALUE" && grandchildXmlNode.firstChild) {
          value += grandchildXmlNode.firstChild.nodeValue;
        }
      }
      if (typeof propNames[pname] == "undefined") {
        // New prop defined on this object.
        propNames[pname] = [xmlNode, perms, owner, value];
        propObject.push(pname);
      } else if (value == "") {
        // Inherited clear property.  The perms and owner might be overridden.
        propNames[pname][1] = perms;
        propNames[pname][2] = owner;
      } else {
        // Inherited overridden property.
        propNames[pname] = [xmlNode, perms, owner, value];
      }
    }
  }
}


function updatePropTable(xmlNode, xmlLeaf, propList, propNames) {
  var contentDiv = document.getElementById("content_properties_yes");
  var id = xmlNode.getAttribute('ID');
  var number = id.replace('obj', '#');
  var leafid = xmlLeaf.getAttribute('ID');
  var leafnumber = leafid.replace('obj', '#');

  var table = newTable("proplist_" + id, ["Property name", "Perms", "Owner", "Value"]);
  var pname, clear, perms, owner, value;
  for (var x = 0; x < propList.length; x++) {
    pname = propList[x];
    clear = propNames[pname][0] !== xmlLeaf;
    perms = propNames[pname][1];
    owner = propNames[pname][2];
    owner = owner.replace('obj', '#');
    owner = parent.parent.TreeFrame.treeGetOrCreateObject(owner, 'inheritance');
    owner = buildObjectLink(owner);
    value = propNames[pname][3];
    if (value.length > 18) {
      value = value.substring(0, 16) + "\u2026"; // That's a "..."
    }

    pname = buildDataLink("." + pname, leafnumber + "." + pname);
    pname.onclick = propClickName;
    if (clear) {
      clear = document.createElement("SPAN");
      clear.className = 'clearProp';
      clear.appendChild(pname);
      pname = clear;
    }
    appendRow(table, [pname, perms, owner, value]);
  }
  var inheritanceTreeNode = parent.parent.TreeFrame.treeGetOrCreateObject(number, 'inheritance');
  var header = document.createElement("div");

  var control = document.createElement("IMG");
  if (propList.length == 0) {
    control.src = "tree-node.gif";
    control.className = "nocontrol";
  } else {
    control.src = "tree-closed.gif";
    control.className = "control";
    control.id = "propcontrol_" + id
    control.onclick = listControlClick;
  }

  header.appendChild(control);
  var headerText = propList.length + " " + (propList.length == 1 ? "property " : "properties ") +
      (xmlNode === xmlLeaf ? "on " : "inherited from ") +
      inheritanceTreeNode.firstChild.title + (propList.length == 0 ? "." : ":");
  var headerspan = document.createElement("span");
  headerspan.appendChild(document.createTextNode(headerText));
  if (propList.length != 0) {
    headerspan.className = "object";
    headerspan.id = "propheader_" + id
    headerspan.onclick = listHeaderClick;
  }
  header.appendChild(headerspan);

  contentDiv.appendChild(header);
  if (propList.length != 0) {
    table.style.display = "none";
    contentDiv.appendChild(table);
    TableSort.init("proplist_" + id);
  }
}

function updateVerbTable(xmlNode, xmlLeaf) {
  var contentDiv = document.getElementById("content_verbs_yes");
  var id = xmlNode.getAttribute('ID');
  var number = id.replace('obj', '#');

  var table = newTable("verblist_" + id, ["Verb name", "Args", "Perms", "Owner"]);
  var childXmlNodeLength = xmlNode.childNodes.length;
  var childXmlNode;
  var vname, args, perms, owner;
  var verbCount = 0;
  for (var x = 0; x < childXmlNodeLength; x++) {
    childXmlNode = xmlNode.childNodes[x];
    if (childXmlNode.nodeName == "VERB") {
      verbCount++;
      args = childXmlNode.getAttribute('DOBJ') + " " +
             childXmlNode.getAttribute('PREP') + " " +
             childXmlNode.getAttribute('IOBJ');
      perms = childXmlNode.getAttribute('PERMS') || "";
      owner = childXmlNode.getAttribute('OWNER');
      owner = owner.replace('obj', '#');
      owner = parent.parent.TreeFrame.treeGetOrCreateObject(owner, 'inheritance');
      owner = buildObjectLink(owner);
      vname = ":";
      var grandchildXmlNodeLength = childXmlNode.childNodes.length;
      var grandchildXmlNode;
      for (var y = 0; y < grandchildXmlNodeLength; y++) {
        grandchildXmlNode = childXmlNode.childNodes[y];
        if (grandchildXmlNode.nodeName == "NAME" && grandchildXmlNode.firstChild) {
          vname += grandchildXmlNode.firstChild.nodeValue;
        }
      }
      vname = buildDataLink(vname, number + vname + "\t" + args);
      vname.onclick = verbClickName;
      appendRow(table, [vname, args, perms, owner]);
    }
  }
  var inheritanceTreeNode = parent.parent.TreeFrame.treeGetOrCreateObject(number, 'inheritance');
  var header = document.createElement("div");

  var control = document.createElement("IMG");
  if (verbCount == 0) {
    control.src = "tree-node.gif";
    control.className = "nocontrol";
  } else {
    control.src = "tree-closed.gif";
    control.className = "control";
    control.id = "verbcontrol_" + id
    control.onclick = listControlClick;
  }

  header.appendChild(control);
  var headerText = verbCount + " " + (verbCount == 1 ? "verb " : "verbs ") +
      (xmlNode === xmlLeaf ? "on " : "inherited from ") +
      inheritanceTreeNode.firstChild.title + (verbCount == 0 ? "." : ":");
  var headerspan = document.createElement("span");
  headerspan.appendChild(document.createTextNode(headerText));
  if (verbCount != 0) {
    headerspan.className = "object";
    headerspan.id = "verbheader_" + id
    headerspan.onclick = listHeaderClick;
  }
  header.appendChild(headerspan);

  contentDiv.appendChild(header);
  if (verbCount != 0) {
    table.style.display = "none";
    contentDiv.appendChild(table);
    TableSort.init("verblist_" + id);
  }
}

function listControlClick() {
  listSetFold(this, this.src.match("tree-closed\.gif$"));
}

function listHeaderClick() {
  var controlId = this.id.replace('header', 'control');
  var control = document.getElementById(controlId);
  listSetFold(control, control.src.match("tree-closed\.gif$"));
}

function listSetFold(node, open) {
  var tableId = node.id.replace('control', 'list');
  var table = document.getElementById(tableId);
  if (table) {
    table.style.display = open ? "block" : "none";
    node.src = open ? "tree-opened.gif" : "tree-closed.gif";
  }
}

function replaceChildNodes(parentNode, newChild) {
  while (parentNode.firstChild) {
    parentNode.removeChild(parentNode.firstChild);
  }
  parentNode.appendChild(newChild);
}

function buildObjectLink(treeNode) {
  var label = document.createElement("SPAN");
  if (treeNode.className == "object") {
    label.appendChild(document.createTextNode(treeNode.firstChild.title));
    label.className = "object";
    label.onclick = objClickName;
  } else {
    label.appendChild(document.createTextNode("<nothing> (#-1)"));
    label.className = "nothing";
  }
  return label;
}

function buildDataLink(dataName, dataDesc) {
  var label = document.createElement("SPAN");
  label.appendChild(document.createTextNode(dataName));
  label.label = dataDesc;
  label.className = "object";
  return label;
}

function objClickName() {
  var number = this.firstChild.data.match(/ \(#([0-9]+)\)$/);
  if (number && number.length == 2) {
    // Look for one specific object.
    number = "obj" + number[1];
    parent.parent.TreeFrame.changeObject(number);
  }  
}

function verbClickName() {
  var m = this.label.match(/^#([0-9]+):([^\t]+)\t([^ ]+) ([^ ]+) ([^ ]+)$/);
  if (m && m.length == 6) {
    // Look for one specific object.
    var number = "obj" + m[1];
    var vname = m[2];
    var dobj = m[3];
    var prep = m[4];
    var iobj = m[5];
    selectDataLink(this);
    loadVerb(number, vname, dobj, prep, iobj);
  }
}

function propClickName() {
  var m = this.label.match(/^#([0-9]+)\.(.+)$/);
  if (m && m.length == 3) {
    // Look for one specific object.
    var number = "obj" + m[1];
    var pname = m[2];
    selectDataLink(this);
    loadProp(number, pname);
  }
}

var selectedNode = null;
function selectDataLink(node) {
  if (selectedNode) {
    selectedNode.className = "object";
  }
  selectedNode = node;
  node.className = "selected";
}

function newTable(id, headers) {
  var table = document.createElement("TABLE");
  table.setAttribute("border", 1);
  table.setAttribute("cellspacing", 0);
  var thead = document.createElement("THEAD");
  var tbody = document.createElement("TBODY");
  table.appendChild(thead);
  table.appendChild(tbody);
  var tr = document.createElement("TR");
  thead.appendChild(tr);
  var th;
  for (var x = 0; x < headers.length; x++) {
    th = document.createElement("TH");
    th.appendChild(document.createTextNode(headers[x]));
    th.setAttribute("label", "nocase");
    tr.appendChild(th);
  }
  table.id = id;
  table.className = "list";
  return table;
}

function appendRow(table, cells) {
  var tr = document.createElement("TR");
  var td;
  for (var x = 0; x < cells.length; x++) {
    td = document.createElement("TD");
    if (cells[x] === null) {
      cells[x] = document.createElement("I").appendChild(document.createTextNode("null"));
    } else if (typeof cells[x] != "object") {
      cells[x] = cells[x].toString() || "\xA0"; // That's an &nbsp;
      cells[x] = document.createTextNode(cells[x]);
    }
    td.appendChild(cells[x]);
    tr.appendChild(td);
  }
  // Add the new row to the table's tbody.
  table.childNodes[1].appendChild(tr);
}

